/*
** Module   : WindowLayout.java
** Abstract : Layout the elements of a window.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140912 Set a new BitmapScreen object when window is resized.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140923 GUI message area layout.
** 006 CA  20141021 Fixes related to window layout during resize.
** 007 MAG 20141021 Clear window borders whenever is resized.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Fixed initial dimension for dynamic windows.
** 011 HC  20141207 Improved window layout algorithm to honor the requested window size.
** 012 HC  20150323 Removed business logic from config classes.
** 013 CA  20150507 Fixed problems related to window sizing (explicit or implicit).
** 014 CA  20150508 Fixed H013 regression when resizing window via mouse (not attributes).
** 015 HC  20150508 Improvements to GUI window scrolling support.
** 016 GES 20150425 Match class name change for GuiOptions to GuiConstants.
** 017 HC  20150612 Implemented support for modal windows.
** 018 HC  20150620 Fixed layout not adding the status area size to the overall window size.
** 019 VIG 20150618 Added menubar layouting.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Layout the elements of a legacy-style window. This includes the window title, the workspace,
 * message area and status area.
 */
public class WindowLayout
implements LayoutManager<GuiOutputManager>
{
   /** Cached GUI driver. */
   private GuiDriver<?,?> gd;
   
   /** Maximum window size. */
   private NativeDimension maxSize;
   
   /**
    * Sets maximum window size.
    * 
    * @param    width
    *           Maximum window width.
    * @param    height
    *           Maximum window height.
    */
   public void setMaxSize(int width, int height)
   {
      maxSize = new NativeDimension(width, height);
   }
   
   /**
    * Get the minimum size of the specified container.
    * 
    * @param    container
    *           The container.
    *          
    * @return   Always {@code null}.
    */
   @Override
   public Dimension minimumSize(Container<GuiOutputManager> container)
   {
      return null;
   }

   /**
    * Place components in specified container.
    * 
    * @param   container
    *          Container to process - must be a window.
    */
   @Override
   public void doLayout(Container<GuiOutputManager> container)
   {
      // TODO: only use pixel/native coordinates, syncing of character values
      //       must be handled by the framework
      
      gd = (GuiDriver<?, ?>) GuiOutputManager.getDriver();
      
      GuiWindow window = (GuiWindow) container;
      
      CoordinatesConversion c = window.coordinates();

      // select renderer
      gd.selectWindow(window.getId().asInt());

      // 0. prerequisites for computing full width/height:
      boolean resizing = window.isResizing();

      // - the message area height
      int mheight = 0;
      MessageAreaGuiImpl marea = window.getGuiMessageArea();
      
      if (marea != null)
         mheight = c.heightToNative(marea.height());
      
      // - the status area height
      int sheight = 0;
      StatusLine<GuiOutputManager> sarea = window.getStatusArea();
      
      if (sarea != null)
         sheight = c.heightToNative(sarea.height()); 

      // - the border delta
      int borderDelta = GuiConstants.BORDER_SIZE * 2;
      
      // - window size
      int    wheight;
      int    wwidth;
      final double cwheight;
      final double cwwidth;
      
      // - workspace size
      NativeDimension workSize = window.getWorkspaceNativeSize();
      
      WindowConfig wc = window.config();
      WindowTitleBar tbar = window.getTitleBar();
      
      if (resizing)
      {
         NativeDimension oldSize     = window.physicalDimension();
         NativeDimension resizedSize = window.getResizedSize();
         
         // - get the new workspace height/width
         workSize.width  = workSize.width + (resizedSize.width - oldSize.width);
         workSize.height = workSize.height + (resizedSize.height - oldSize.height);
         
         int minmaxWidthDelta = 0;
         int minmaxHeightDelta = 0;
         
         // limit the new width by the config min/max width
         if (workSize.width < wc.minWidthPixels)
         {
            minmaxWidthDelta = wc.minWidthPixels - workSize.width;
            workSize.width = wc.minWidthPixels;
         }
         else if (workSize.width > wc.maxWidthPixels)
         {
            minmaxWidthDelta = wc.maxWidthPixels - workSize.width;
            workSize.width = wc.maxWidthPixels;
         }
         
         // limit the new height by the config min height
         if (workSize.height < wc.minHeightPixels)
         {
            minmaxHeightDelta = wc.minHeightPixels - workSize.height;
            workSize.height = wc.minHeightPixels;
         }
         else if (workSize.height > wc.maxHeightPixels)
         {
            minmaxHeightDelta = wc.maxHeightPixels - workSize.height;
            workSize.height = wc.maxHeightPixels;
         }
         
         wheight  = resizedSize.height + minmaxHeightDelta;
         wwidth   = resizedSize.width + minmaxWidthDelta;
         cwheight = c.heightFromNative(wheight);
         cwwidth  = c.widthFromNative(wwidth);

         // - set the titlebar's width
         double ctbwidth = c.widthFromNative(wwidth - borderDelta);
         tbar.setWidth(ctbwidth);
         tbar.doLayout();
      }
      else
      {
         // - layout the titlebar now, is needed to compute the window height
         double ctbwidth = c.widthFromNative(workSize.width);
         tbar.setWidth(ctbwidth);
         tbar.doLayout();
         int tbheight = c.heightToNative(tbar.height());

         wheight  = workSize.height + tbheight + mheight + sheight + borderDelta; 
         wwidth   = workSize.width + borderDelta;
         
         if (maxSize != null)
         {
            // adjust to maximum window size if needed
            
            int deltaW = Math.max(0, wwidth - maxSize.width);
            int deltaH = Math.max(0, wheight - maxSize.height);
            
            wwidth =- deltaW;
            wheight =- deltaH;
            
            workSize.width =- deltaW;
            workSize.height =- deltaH;
         }
         
         cwheight = c.heightFromNative(wheight);
         cwwidth  = c.widthFromNative(wwidth); 
      }
      
      window.setWorkspaceNativeSize(workSize.width, workSize.height);
      
      int tbheight = c.heightToNative(tbar.height());
      
      // TODO: in addition to MIN/MAX-HEIGHT attrs the window seems to have
      //       physical size limits - for width it is a minimum window title bar plus borders
      //       and for height it is the title bar height plus message area plus window status bar
      //       plus borders.
      
      if (resizing)
      {
         // clear window border
         clearBorder(gd, new NativeDimension(wwidth, wheight));
      }
      else
      {
         gd.resizeWindow(wwidth, wheight);
      }

      window.setSize(cwwidth, cwheight);
      
      // update screen bitmap
      GuiOutputManager.instance().setBitmap(window.getId().asInt(), new ScreenBitmap(wheight, wwidth));
      
      // all coordinates for the following widgets are relative to the OS window (the container)
      int x = GuiConstants.BORDER_SIZE;
      int y = GuiConstants.BORDER_SIZE;
      
      // 1. set the title into position:
      tbar.setPhysicalLocation(x, y);
      tbar.setLocation(c.widthFromNative(x), c.heightFromNative(y));
      
      // 2. set the window workarea (this is the legacy 4GL window, where frames are placed)
      // - bellow the title and above the message area
      // - its size does not include the borders
      // - its size needs to be computed first, as the general window size depends on it
      y = y + tbheight;
      
      if (window instanceof WindowGuiImpl)
      {
         Menu<GuiOutputManager> menubar = ((WindowGuiImpl)window).getMenubar();
         if (menubar != null)
         {
            MenuConfig mc = menubar.config();
            menubar.setPhysicalLocation(x, y);
            menubar.setLocation(c.widthFromNative(x), c.heightFromNative(y));
            menubar.setWidth(cwwidth);
            mc.widthChars = cwwidth;
            mc.widthPixels = wwidth;
            menubar.doLayout();
            y = y + 1;
         }
      }
      
      window.setWorkspaceLocation(x, y);
      
      // the y position for the next UI element
      y = y + workSize.height;
      
      // 3. set the message area
      if (marea != null)
      {
         marea.setPhysicalLocation(x, y);
         marea.setLocation(c.widthFromNative(x), c.heightFromNative(y)); 
         marea.doLayout();
      }
      
      // 4. set the status area
      if (marea != null)
      {
         y = y + mheight;
      }
      
      if (sarea != null)
      {
         window.getStatusArea().setPhysicalLocation(x, y);
         window.getStatusArea().setLocation(c.widthFromNative(x), c.heightFromNative(y));
      }
   }

   /**
    * Whenever window is resized the right and bottom borders became dirty.  In order to prevent
    * dirty borders right and bottom borders are cleaned.
    * 
    * @param   gd
    *          GUI device.
    * @param   d
    *          Window dimension.
    */
   private void clearBorder(GuiDriver gd, NativeDimension d)
   {
      // main window border size
      int border = GuiConstants.BORDER_SIZE;
      
      // clear left and bottom borders
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
      gd.fillRect(d.width - border, 0, border, d.height);
      gd.fillRect(0, d.height - border, d.width, border);      
   }
}
